# -*- coding: utf-8 -*-
"""
Created on Sat May 11 09:22:54 2019

"""

## Example of Semi-Parametric Estimation
##
## Inspired by blogs and Markow Wanderer code


## House Keepingin Python
## %reset  # works in iPython to clear the workspace, etc.
## %reset -f   #no questios asked...

import numpy as np
from sklearn.datasets import make_spd_matrix
import math
import statsmodels.api as sm # least squares in Python
from sklearn.ensemble import RandomForestRegressor 


# Set up the environment
# -------------------------------------------------------------
randomseednumber = 43 
np.random.seed(randomseednumber)


N  = 100 # No. obs
k  = 30  # = No. variables in x_i
theta = 0.5 # Structural parameter
b= [1/k for k in range(1,k+1)] # x weights
sigma = make_spd_matrix(k,randomseednumber) #


## Define the functional form of the g(x) and m(x)
# -------------------------------------------------------------
def g(x):
    #return np.power(np.sin(x),2)
    return np.power(np.cos(x),2)

def m(x,nu=0.,gamma=1.):
    #return 0.5/math.pi*(np.sinh(gamma))/(np.cosh(gamma)-np.cos(x-nu))
    return np.sin(x) + np.cos(x)



# -------------------------------------------------------------
# NUmber of simulations
ndraws = 100

# Array of estimated thetas to store results
theta_est = np.zeros((ndraws, 3))
theta_1   = np.zeros((ndraws, 1))
theta_2   = np.zeros((ndraws, 1))

maxdepth = 3 # max depth of a tree

for i in range(ndraws):
    
    if ((i % 100) == 0):
        print(i)
    
    # Generate data: no. obs x no. variables in x_i
    X = np.random.multivariate_normal(np.ones(k),sigma,size=[N,])
    G = g(X @ b)
    M = m(X @ b)
    D = M+np.random.standard_normal(size=[N,])
    Y = np.dot(theta,D)+G+np.random.standard_normal(size=[N,])
    #
  
    # OLS and  Make it "less Naive"
    # ------------------------------------------------------
    DX = np.concatenate((D.reshape(N,1), X, np.ones([N,1])), axis=1) # [X D ones(N,1)]
    OLS_DX = sm.OLS(Y, DX)
    results = OLS_DX.fit()
    theta_est[i][0] = results.params[0]

    # Naive double machine Learning without sample-splitting and cross-fitting
    naiveDMLg =RandomForestRegressor(max_depth=maxdepth)
    
    # Compute ghat
    naiveDMLg.fit(X,Y)
    G_hat = naiveDMLg.predict(X)
    naiveDMLm =RandomForestRegressor(max_depth=maxdepth)
    naiveDMLm.fit(X,D)
    M_hat = naiveDMLm.predict(X)
 
    # vhat as residual
    V_hat = D-M_hat
    theta_est[i][1] = np.mean(np.dot(V_hat,Y-G_hat))/np.mean(np.dot(V_hat,D))

    # Cross-fitting DML -----------------------------------
    
    # Split the sample
    I = np.random.choice(N,np.int(N/2),replace=False)
    I_C = [x for x in np.arange(N) if x not in I]
    
    # Ghat for both
    Ghat_1 = RandomForestRegressor(max_depth=maxdepth).fit(X[I],Y[I]).predict(X[I_C])
    Ghat_2 = RandomForestRegressor(max_depth=maxdepth).fit(X[I_C],Y[I_C]).predict(X[I])
    
    # Mhat and vhat for both
    Mhat_1 = RandomForestRegressor(max_depth=maxdepth).fit(X[I],D[I]).predict(X[I_C])
    Mhat_2 = RandomForestRegressor(max_depth=maxdepth).fit(X[I_C],D[I_C]).predict(X[I])
    
    Vhat_1 = D[I_C]-Mhat_1
    Vhat_2 = D[I] - Mhat_2
    
    theta_1[i] = np.mean(np.dot(Vhat_1,(Y[I_C]-Ghat_1)))/np.mean(np.dot(Vhat_1,D[I_C]))
    theta_2[i] = np.mean(np.dot(Vhat_2,(Y[I]-Ghat_2)))/np.mean(np.dot(Vhat_2,D[I]))
    
    theta_est[i][2] = 0.5*(theta_1[i]+theta_2[i])
    
# Visualize    
import matplotlib
import seaborn as sns # let's try to use seaborn, as long as it creates valid matplotlibs
import matplotlib.pyplot as plt
from scipy import stats

sns.set_style("white")
matplotlib.rcParams['text.usetex'] = False # True unleashes HELL but then works



## Create a nice picture to plot for SLIDES
## ---------------------------------------------------------------------
sns_plot = sns.distplot((theta_1+theta_2)/2, hist=False, label="Double ML + cross fitting");
sns_plot = sns.distplot(theta_est[:][0], hist=False, label="OLS");
sns_plot = sns.distplot(theta_est[:][1], hist=False, label="`Naive' ML");
plt.axvline(0.5, 0, 3, color='red', linewidth=1)
sns_plot.spines['right'].set_visible(False)
sns_plot.spines['top'].set_visible(False)

#plt.xlabel(r'$\widehat{\theta}$',fontsize=12,usetex=True) # This Messes UP the PS renderer... (?)
plt.xlabel(r'$\theta$',fontsize=12) # This Messes UP the PS renderer... (?)
plt.ylabel('Frequency', fontsize=12)

#sns_plot.figure.savefig("double_ml.eps")




