# -*- coding: utf-8 -*-
"""
Created on Tue May 14 17:26:10 2019

@author: Michal Andrle (mandrle@imf.org)
"""


## imports
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LassoCV

import statsmodels.api as sm

## Illustration of Double-Selection LASSO and post-lasso 


# Set up the simulation
N = 100
k = 90

theta = 0.5 # The low-dimensional parameter to estimate...


# generate the data and the model

beta0 = np.zeros([k,1])
for i in np.arange(1,k+1):
    beta0[i-1] = (1/i)**2
    
beta1 = 0.7*beta0
    

## ----------------------------------------------------------------------------
## Create the Data-Generating Process [DGP] (normally you'd load data...)
## ----------------------------------------------------------------------------
def generate_data():
    # co-variance of X
    SIG = np.zeros([k,k])
    for ix_r in np.arange(k):
        for ix_c in np.arange(k):
            SIG[ix_r, ix_c] = 0.5 **np.absolute(ix_r - ix_c)

    # Sample the X explanatory variables 
    X = np.random.multivariate_normal(np.zeros(k), SIG, size=[N,])


    # Generate the Sample
    D = np.dot(X,beta0) + np.random.standard_normal(size=[N,1])
    Y = theta*D + np.dot(X, beta1) + np.random.standard_normal(size=[N,1])
    
    return (Y, D, X)


## ----------------------------------------------------------------------------
## Estimation using double-selection post-lasso
## ----------------------------------------------------------------------------
def double_selection(Y, D, X, debug=False):


    m1 = LassoCV(cv=5, random_state=0, max_iter=200, tol=0.01)
    m2 = LassoCV(cv=5, random_state=0, max_iter=200, tol=0.01)

    m1.fit(X, Y.ravel())
    m2.fit(X, D.ravel())

    # Selecth thresholded regressors [coef_ > 0.01]
    coef_threshold = 0.0

    ix1 = m1.coef_ > coef_threshold
    ix2 = m2.coef_ > coef_threshold

    ix = ix1 | ix2  # create the union of the variables to work with " | is or"

    # Select the variables to RUN OLS with
    X_new = X[:, ix]
    
    if debug:
        print('Number of selected variables: %s' % np.sum(ix))
        
    return X_new


## ----------------------------------------------------------------------------
## Estimate Least Squares with the [D X_new]
## ----------------------------------------------------------------------------
def post_lasso(Y, D, X_new):

    XOLS = np.hstack((D, X_new)) # concatenate D with X_new for the OLS

    #final_ols = LinearRegression()
    #final_ols.fit(XOLS, Y)
    
    final_ols = sm.OLS(Y, XOLS).fit()
    
    #return final_ols.coef_[0]
    return final_ols.params



##############################################################################
## RUNNING THE SAMPLING
##############################################################################
ndraws = 1000

theta_est = np.zeros(ndraws,)
theta_ols = np.zeros(ndraws,)

for i in np.arange(ndraws):
    
    if ((i % 100) == 0):
        print('Iteration #: %d ' % i)
        
    # resample the data
    Y, D, X = generate_data()
    
    # double-selection using cross-validated LASSO
    X_new = double_selection(Y, D, X)
    
    # OLS using the union of double-selected variables 'X_new'
    est_coef = post_lasso(Y, D, X_new)
    theta_est[i] = est_coef[0]

    # Compare to FULL OLS    [only when it's feasible, i.e. k < N]
    if (k < N):
        ols = LinearRegression()
        ols.fit(np.hstack((D, X)), Y)
        theta_ols[i] = ols.coef_[0][0]

    
    
# Create the plot of theta_est distribution
# -------------------------------------------------------------------
import seaborn as sns # let's try to use seaborn, as long as it creates valid matplotlibs
sns.set_style("white")
sns_plot = sns.distplot(theta_est-theta, hist=False, label="Double Selection");
if (k < N):
    sns_plot = sns.distplot(theta_ols-theta, hist=False, label="OLS [no selection]");

plt.axvline(theta-theta, 0, 4, color='red', linewidth=1)
sns_plot.spines['right'].set_visible(False)
sns_plot.spines['top'].set_visible(False)

plt.xlabel(r'$\hat\theta - \theta$',fontsize=12) # This Messes UP the PS renderer... (?)
plt.ylabel('Frequency', fontsize=12)

#sns_plot.figure.savefig("post_double_selection_ols.eps")

























