# -*- coding: utf-8 -*-
"""
Created on Tue May 21 17:04:30 2019

@author: MAndrle@imf.org
"""


## imports
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegressionCV
from sklearn.linear_model import LassoCV

import statsmodels.api as sm


"""
This simple code create a linear model with a binary treatment variable
and tests post-double-selection to estimate it with the true specification
uknown.

Y(i) = 0.5 * D(i) + beta1*X5 + beta2*X13 + beta3*X17 + e1

D(i) = 0  if  prob_D < 0.5
     = 1      prob_D => 0.5


prob_D(i) = logit(gamma1*X5 + gamma2*X7 + gamma3*X13 + gamma4*X20 + e2)     

"""





# Model Setup
N = 100
k = 80

beta1 = 0.3
beta2 = 0.7
beta3 = 1.2

gamma1 = 0.2
gamma2 = 0.7
gamma3 = -0.7
gamma4 = 0.65

beta_true = 2.0


## Data-generating Process
def dgp():
   
    SIG = np.zeros([k,k])
    for ix_r in np.arange(k):
        for ix_c in np.arange(k):
            SIG[ix_r, ix_c] = 0.5 **np.absolute(ix_r - ix_c)
    
    # Sample the X explanatory variables 
    X = np.random.multivariate_normal(np.zeros(k), SIG, size=[N,])
    #X = np.random.randn(N, k) # these are not correlated so far... [so easy]
    
    ez = np.random.randn(N, )
    Z = 2 + gamma1 * X[:, 4] + gamma2 * X[:, 6] + gamma3*X[:, 12] + gamma4*X[:, 19] + ez
    prob = np.exp(Z) /(1 + np.exp(Z))
    ix_treat = (prob >= 0.5)

    # create the [0|1] treatment variable
    D = np.zeros((N,1))
    D[ix_treat] = 1
    
    # flatten the thing... aaaaargh
    D.reshape(N,)

    ey = 0.3 * np.random.randn(N,)
    Y  =  beta_true * D.reshape(N,) + beta1*X[:, 4] + beta2*X[:, 12] + beta3*X[:, 16] + ey

    return (Y, D, X)


## Estimation using post-lasso
__debug = False


S = 1000 # number of samples to draw


# pre-allocate the storage for the sampling...
beta       = np.zeros((S,1))
beta_ols   = np.zeros((S,1))
beta_nocov = np.zeros((S,1))

for s in np.arange(S):
    
    # inform about the progress
    if (s % 50 == 0):
        print(s)
    
    
    # Sample from the DGP
    (Y, D, X) = dgp()
    
        
    # Y <- X [direct leg]
    m_yx = LassoCV(cv=4, random_state=0, max_iter=200, tol=0.01)
    m_yx.fit(X, Y)
    
    
    # D <- X [direct leg] [but now logistic regression]
    m_dx = LogisticRegressionCV(cv=4, random_state=0, penalty="l1", solver="liblinear")
    m_dx.fit(X,D.flatten(order='C'))
    
    
    # Select the union of the models
    ix1  = m_yx.coef_ > 0.01
    ix2_ = m_dx.coef_ > 0.01
    ix2  = ix2_.reshape(len(ix2_.T), ) # this is sooooooo stupid...
    
    # Union of the selected indices [logigal or]
    ix_sel = ix1 | ix2
    
    if __debug:
        print(np.arange(k)[ix1 | ix2])
        
    # New set of variables to do post-lasso with
    X_SEL = X[:, ix_sel]
    
    

    # OLS -- post-double-selection lasso
    DX_SEL = np.hstack((D, X_SEL)) # this is new set of co-variates
    ols = sm.OLS(Y, DX_SEL).fit()
    beta[s] = ols.params[0]
    
    # OLS with no variable selection
    ols_base = sm.OLS(Y, np.hstack((D, X)) ).fit()
    beta_ols[s] = ols_base.params[0]
    
    ## FULL OLS -- no covariates
    ols_nocov = sm.OLS(Y, D).fit()
    beta_nocov[s] = ols_nocov.params[0]
    
    

# Do the histograms
# -----------------------------------------------------------------
import seaborn as sns # let's try to use seaborn, as long as it creates valid matplotlibs
sns.set_style("white")



sns_plot = sns.distplot(beta, hist=False, color='red',label="Double Selection");
sns_plot = sns.distplot(beta_nocov, hist=False, color='orange', label="OLS [no covariates]");
if (k < N):
    sns_plot = sns.distplot(beta_ols, hist=False, label="OLS [no selection]");

plt.axvline(beta_true, 0, 4, color='gray', linewidth=1)
   
sns_plot.spines['right'].set_visible(False)
sns_plot.spines['top'].set_visible(False)


sns_plot.figure.savefig("post_double_selection_treatment.eps")


## For the Last Sample, show the standard analysis of confidence bands
# -----------------------------------------------------------------

z_fig, ax = plt.subplots(figsize=(7,4))

bar_labels = ('Truth','OLS','No cov.','Double-Selection')
y_pos = np.arange(len(bar_labels))

y_means = [2, beta_ols[-1], beta_nocov[-1], beta[-1]];  # note that x[-1] picks the last element...

y_errors = np.array([0, -ols_base.bse[0], -ols_nocov.bse[0], -ols.bse[0]])
y_errors = np.vstack((-y_errors, -y_errors))

the_bar = plt.bar(y_pos, [2, beta_ols[-1], beta_nocov[-1], beta[-1]], yerr = 2*y_errors);
plt.xticks(y_pos, bar_labels, fontsize=12);

ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
    
z_fig.savefig("post_dsel_treatment_bars.eps")






























